package com.galaxis.manatee.service;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.galaxis.capsule.bo.ChuanyunObject;
import com.galaxis.capsule.dto.MaterialDTO;
import com.galaxis.manatee.constant.ChuanYunConstant;
import com.galaxis.manatee.constant.DingTalkConstant;
import com.galaxis.manatee.constant.StringConstant;
import com.galaxis.manatee.dao.*;
import com.galaxis.manatee.entity.chuanyun.data.object.*;
import com.galaxis.manatee.entity.chuanyun.dto.*;
import com.galaxis.manatee.entity.chuanyun.bo.MaterialResendBO;
import com.galaxis.manatee.entity.ding.DingTalkProcessInstance;
import com.galaxis.manatee.manager.ChuanYunManager;
import com.galaxis.manatee.manager.ClawFeign;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.domain.Example;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import java.math.BigDecimal;
import java.util.*;

/**
 * @author zcj
 * @version 0.1
 * @date 2020/2/13 11:53 上午
 */
@Service
@Slf4j
public class ChuanyunScheduledTask {

    private final Integer pageSize=20;
    private final ChuanYunManager chuanYunManager;
    private final ChuanyunReimbursementDao chuanyunReimbursementDao;
    private final ChuanyunProjectDao chuanyunProjectDao;
    private final ChuanyunMaterialDao chuanyunMaterialDao;
    private final DingTalkProcessInstanceDao dingTalkProcessInstanceDao;
    private final ChuanyunCompanyDao chuanyunCompanyDao;
    private final ChuanyunSupplierDao chuanyunSupplierDao;
    private final ChuanyunGroupProjectDao chuanyunGroupProjectDao;
    private final ChuanyunUserCompanyDao chuanyunUserCompanyDao;
    private final ChuanyunUserDao chuanyunUserDao;
    private final ChuanyunCostDao chuanyunCostDao;
    private final ChuanyunReimbursementSubjectDao chuanyunReimbursementSubjectDao;
    private final ClawFeign clawFeign;

    public ChuanyunScheduledTask(ChuanYunManager chuanYunManager, ChuanyunReimbursementDao chuanyunReimbursementDao, ChuanyunProjectDao chuanyunProjectDao, ChuanyunMaterialDao chuanyunMaterialDao, DingTalkProcessInstanceDao dingTalkProcessInstanceDao, ChuanyunCompanyDao chuanyunCompanyDao, ChuanyunSupplierDao chuanyunSupplierDao, ChuanyunGroupProjectDao chuanyunGroupProjectDao, ChuanyunUserCompanyDao chuanyunUserCompanyDao, ChuanyunUserDao chuanyunUserDao, ChuanyunCostDao chuanyunCostDao, ChuanyunReimbursementSubjectDao chuanyunReimbursementSubjectDao, ClawFeign clawFeign) {
        this.chuanYunManager = chuanYunManager;
        this.chuanyunReimbursementDao = chuanyunReimbursementDao;
        this.chuanyunProjectDao = chuanyunProjectDao;
        this.chuanyunMaterialDao = chuanyunMaterialDao;
        this.dingTalkProcessInstanceDao = dingTalkProcessInstanceDao;
        this.chuanyunCompanyDao = chuanyunCompanyDao;
        this.chuanyunSupplierDao = chuanyunSupplierDao;
        this.chuanyunGroupProjectDao = chuanyunGroupProjectDao;
        this.chuanyunUserCompanyDao = chuanyunUserCompanyDao;
        this.chuanyunUserDao = chuanyunUserDao;
        this.chuanyunCostDao = chuanyunCostDao;
        this.chuanyunReimbursementSubjectDao = chuanyunReimbursementSubjectDao;
        this.clawFeign = clawFeign;
    }

    /**
     * 定时将氚云中的数据保存到manatee
     */
    @Scheduled(fixedDelay = 43200000L)
    private void getDataFromChuanyun(){
        //获取项目数据
        getProjectInformationFromChuanyun();
    }

    /**
     * 定时将物料信息从氚云中同步
     */
    @Scheduled(fixedDelay = 27600000L)
    private void materialInformationSync(){
        var pageSize=100;
        var start=0;
        var totalCount=0L;
        var flag=true;
        while (flag){
            try {
                //从氚云查询数据
                var filter=Filter.instance(start,start+pageSize,true);
                var chuanyunFindAllResponse=chuanYunManager.findAll(ChuanyunMaterialDO.SCHEMA_CODE,filter);
                if(chuanyunFindAllResponse.getReturnData()==null){
                    flag=false;
                    continue;
                }
                if(chuanyunFindAllResponse.getReturnData().getTotalCount()>=start+pageSize){
                    start+=pageSize;
                }else{
                    flag=false;
                }
                //转化为POJO
                var objectMapper=new ObjectMapper();
                List<ChuanyunMaterialDO> result=objectMapper.convertValue(chuanyunFindAllResponse.getReturnData().getBizObjectArray(), new TypeReference<>(){});
                //保存
                result.removeIf(chuanyunMaterialDO -> chuanyunMaterialDao.findById(chuanyunMaterialDO.getObjectId()).isPresent());
                chuanyunMaterialDao.saveAll(result);
                log.info("保存物料信息"+result.size());
                totalCount+=result.size();
            }catch (Exception e){
                log.error(e.getMessage());
                e.printStackTrace();
            }
        }
        log.info("物料信息保存保存"+totalCount);
        //将氚云数据同步到mysql之后将U9的数据同步到氚云
        getMaterialFromU9();
    }

    @Scheduled(fixedDelay = 27600000L,initialDelay = 10000L)
    private void reimbursementInformationSync(){
        getReimbursementFromChuanyun();
    }

    /**
     * 将U9中物料信息更新到氚云
     */
    private void getMaterialFromU9(){
        int page=1;
        int size=50;
        var totalAmount=0;
        var totalPage=Integer.MAX_VALUE;
        Map<String,String> parameters=new HashMap<>(2);
        parameters.put("page", Integer.toString(page));
        parameters.put("size", Integer.toString(size));

        while (page<=totalPage){
            List<ChuanyunMaterialDO> toSave=new ArrayList<>();
            //获取U9中的物料数据
            Page<MaterialDTO> result=clawFeign.findMaterialPageable(parameters);
            totalPage=result.getTotalPages();
            List<ChuanyunMaterialDO> chuanyunMaterialList=new ArrayList<>();
            //遍历获取的数据，如果系统中已经存在就不进行保存
            result.getContent().forEach(materialDTO -> {
                List<ChuanyunMaterialDO> list=chuanyunMaterialDao.findAllByCode(materialDTO.getCode());
                if(list.size()==0){
                    ChuanyunMaterialDO chuanyunMaterialDO=ChuanyunMaterialDO.fromMaterialDTO(materialDTO);
                    chuanyunMaterialList.add(chuanyunMaterialDO);
                    toSave.add(chuanyunMaterialDO);
                }
            });
            chuanyunMaterialDao.saveAll(chuanyunMaterialList);
            totalAmount+=chuanyunMaterialList.size();
            page++;
            parameters.put("page",Integer.toString(page));
            log.info("共"+totalPage+"页，当前第"+page+"页，更新了"+chuanyunMaterialList.size()+"条数据");
            //将更新的的物料信息同步到氚云
            getMaterialFromU9(toSave);
        }
        log.info("更新U9物料数据结束,共更新数据"+totalAmount);
    }

    /**
     * 用于将需要更新的数据同步到氚云
     * @param toSave    待保存的数据集合
     */
    private void getMaterialFromU9(List<ChuanyunMaterialDO> toSave){
        ObjectMapper objectMapper=new ObjectMapper();
        List<String> toSaveString=new ArrayList<>();
        toSave.forEach(chuanyunMaterialDO -> {
            try {
                toSaveString.add(objectMapper.writeValueAsString(chuanyunMaterialDO));
            } catch (JsonProcessingException e) {
                log.info(chuanyunMaterialDO+"");
                e.printStackTrace();
            }
        });
        //将新增加的数据更新到氚云
        chuanYunManager.saveAll(ChuanyunMaterialDO.SCHEMA_CODE,toSaveString,true);
        log.info("物料信息更新到氚云"+toSave.size());
    }

    /**
     * 定时更新项目信息
     */
    private void getProjectInformationFromChuanyun(){
        var start=0;
        var totalCount=0L;
        var flag=true;
        while (flag){
            try {
                //从氚云查询数据
                var filter=Filter.instance(start,start+pageSize,true);
                var chuanyunFindAllResponse=chuanYunManager.findAll(ChuanyunProjectDO.SCHEMA_CODE,filter);
                if(chuanyunFindAllResponse.getReturnData()==null){
                    flag=false;
                    continue;
                }
                if(chuanyunFindAllResponse.getReturnData().getTotalCount()>=start+pageSize){
                    start+=pageSize;
                }else{
                    flag=false;
                }
                //转化为POJO
                var objectMapper=new ObjectMapper();
                List<ChuanyunProjectDO> result=objectMapper.convertValue(chuanyunFindAllResponse.getReturnData().getBizObjectArray(), new TypeReference<>(){});
                //更新缓存中的项目映射表
                result.forEach(projectBO -> ChuanYunConstant.PROJECT_CODE.put(projectBO.getProjectCode(),projectBO.getObjectId()));
                totalCount=chuanyunFindAllResponse.getReturnData().getTotalCount();
                //保存
                result.forEach(chuanyunProjectDO -> {
                    var tmpProjectDO=chuanyunProjectDao.findById(chuanyunProjectDO.getObjectId());
                    tmpProjectDO.ifPresent(projectDO -> chuanyunProjectDO.setIsCreate(projectDO.getIsCreate()));
                    //将projectManger和projectOwner替换为主键
                    if(!StringUtils.isEmpty(chuanyunProjectDO.getProjectManager())){
                        var projectManagerIds=new StringBuilder();
                        for (ChuanyunObject object:chuanyunProjectDO.getProjectManagerObject()){
                            projectManagerIds.append(object.getObjectId()).append(StringConstant.SEMICOLON);
                        }
                        chuanyunProjectDO.setProjectManager(projectManagerIds.toString());
                    }
                    if(!StringUtils.isEmpty(chuanyunProjectDO.getProjectOwner())){
                        chuanyunProjectDO.setProjectOwner(chuanyunProjectDO.getProjectOwnerObject().getObjectId());
                    }
                });
                chuanyunProjectDao.saveAll(result);
            }catch (Exception e){
                log.error(e.getMessage());
                e.printStackTrace();
            }
        }
        log.info("项目信息保存保存"+totalCount+"项目映射更新完成");
    }

    /**
     * 获取报销单
     */
    @Transactional(rollbackFor = Exception.class)
    protected void getReimbursementFromChuanyun(){
        var objectMapper=new ObjectMapper();
        var start=0;
        var totalCount=0L;
        var flag=true;
        List<String> matchersString=new ArrayList<>();
        matchersString.add("Status_2,1");
        while (flag){
            try {
                //从氚云查询数据
                var filter= Filter.instance(start,start+pageSize,true,Filter.AND,matchersString);
                var chuanyunFindAllResponse=chuanYunManager.findAll(ChuanyunReimbursementDO.SCHEMA_CODE,filter);
                if(chuanyunFindAllResponse.getReturnData()==null){
                    flag=false;
                    continue;
                }
                if(chuanyunFindAllResponse.getReturnData().getTotalCount()>=start+pageSize){
                    start+=pageSize;
                }else{
                    flag=false;
                }
                //转化为POJO
                List<ChuanyunReimbursementDO> result=objectMapper.convertValue(chuanyunFindAllResponse.getReturnData().getBizObjectArray(), new TypeReference<>(){});
                //获取项目总数
                totalCount=chuanyunFindAllResponse.getReturnData().getTotalCount();
                //保存
                result.removeIf(chuanyunReimbursementDO->chuanyunReimbursementDao.findById(chuanyunReimbursementDO.getObjectId()).isPresent());
                result.forEach(chuanyunReimbursementDO -> {
                    //将报销单中关联信息的id转化为具体的数据
                    try{
                        ChuanyunCompanyDO chuanyunCompanyDO=chuanyunCompanyDao.findById(chuanyunReimbursementDO.getBeneficiaryId()).orElseThrow();
                        if("对公".equals(chuanyunReimbursementDO.getPrivateOrPublic())&&!StringUtils.isEmpty(chuanyunReimbursementDO.getSupplierId())){
                            ChuanyunSupplierDO chuanyunSupplierDO=chuanyunSupplierDao.findById(chuanyunReimbursementDO.getSupplierId()).orElseThrow();
                            chuanyunReimbursementDO.setSupplierCode(chuanyunSupplierDO.getSupplierCode());
                        }
                        StringBuilder chuanyunGroupProjectCode=new StringBuilder();
                        if(!StringUtils.isEmpty(chuanyunReimbursementDO.getGroupProjectId())){
                            ChuanyunGroupProjectDO chuanyunGroupProjectDO=chuanyunGroupProjectDao.findById(chuanyunReimbursementDO.getGroupProjectId()).orElseThrow();
                            chuanyunReimbursementDO.setGroupProjectCode(chuanyunGroupProjectDO.getProjectCode());
                            chuanyunGroupProjectCode.append(chuanyunGroupProjectDO.getProjectCode());
                        }
                        chuanyunReimbursementDO.setReimburseUserId(chuanyunReimbursementDO.getReimburseUserObject().getObjectId());
                        ChuanyunUserDO chuanyunUserDO=chuanyunUserDao.findById(chuanyunReimbursementDO.getReimburseUserId()).orElseThrow();
                        List<ChuanyunUserCompanyDO> userCompanyList=chuanyunUserCompanyDao.findAllByUserId(chuanyunReimbursementDO.getReimburseUserId());
                        if(userCompanyList.size()!=1){
                            log.error(chuanyunUserDO.getName()+chuanyunReimbursementDO.getReimburseUserId());
                        }else{
                            ChuanyunUserCompanyDO chuanyunUserCompany=userCompanyList.get(0);
                            chuanyunReimbursementDO.setReimburseUserCostCode(chuanyunUserCompany.getCostCode());
                        }
                        ChuanyunCostDO chuanyunCostDO=chuanyunCostDao.findById(chuanyunReimbursementDO.getBeneficiaryCostId()).orElseThrow();
                        chuanyunReimbursementDO.setBusinessDate(chuanyunReimbursementDO.getCreatedTime());
                        chuanyunReimbursementDO.setReimburseUserCompanyCode(chuanyunCompanyDO.getErpCode());
                        chuanyunReimbursementDO.setReimburseUserCode(chuanyunUserDO.getEmployeeNumber());
                        chuanyunReimbursementDO.setBeneficiaryBankCode(chuanyunCompanyDO.getBankCode());
                        chuanyunReimbursementDO.setBeneficiaryBankAccount(chuanyunCompanyDO.getBankAccount());
                        chuanyunReimbursementDO.setBeneficiaryCostCode(chuanyunCostDO.getCostCode());
                        chuanyunReimbursementDO.setCurrency("C001");
                        chuanyunReimbursementDO.setExchangeRate(new BigDecimal("1"));
                        chuanyunReimbursementDO.setSettlementMethod("101");
                        chuanyunReimbursementDO.setErpFlag(0);
                        chuanyunReimbursementDO.getDetailList().forEach(chuanyunReimbursementDetailDO -> {
                            ChuanyunReimbursementSubjectDO chuanyunReimbursementSubjectDO=chuanyunReimbursementSubjectDao.findById(chuanyunReimbursementDetailDO.getReimburseSubjectId()).orElseThrow();
                            chuanyunReimbursementDetailDO.setReimburseSubjectCode(chuanyunReimbursementSubjectDO.getSubjectCode());
                            chuanyunReimbursementDetailDO.setReimburseSubjectName(chuanyunReimbursementSubjectDO.getSubjectName());
                            chuanyunReimbursementDetailDO.setReimburseUserCompanyCode(chuanyunCompanyDO.getErpCode());
                            chuanyunReimbursementDetailDO.setGroupProjectCode(chuanyunGroupProjectCode.toString());
                            chuanyunReimbursementDetailDO.setReimburseUserCostCode(chuanyunCostDO.getCostCode());
                        });
                    }catch (Exception e){
                        log.warn(e.getMessage());
                        e.printStackTrace();
                    }
                });
                chuanyunReimbursementDao.saveAll(result);
            }catch (Exception e){
                log.error(e.getMessage());
                e.printStackTrace();
            }
        }
        log.info("报销信息保存保存"+totalCount);
    }

    /**
     * 将物料补发数据同步到氚云
     * 预计2020年4月份之后逐步将各个部门物料补发数据迁移到氚云进行审批，这个功能逐步废弃
     * 不能调用这个方法！！！！！
     */
    @Deprecated
    private void putMaterialResendInstance(){
        var currentPage=0;
        var totalPage=Integer.MAX_VALUE;

        var objectMapper=new ObjectMapper();
        objectMapper.registerModule(new JavaTimeModule());
        objectMapper.disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);

        while (currentPage<totalPage){
            List<String> strings=new ArrayList<>();
            PageRequest pageRequest=PageRequest.of(currentPage,pageSize);
            DingTalkProcessInstance dingTalkProcessInstance=new DingTalkProcessInstance();
            dingTalkProcessInstance.setProcessInstanceId(DingTalkConstant.PROCESS_CODE_MATERIAL_RESEND);
            dingTalkProcessInstance.setUpdatedInChuanyun(false);
            Page<DingTalkProcessInstance> pageList=dingTalkProcessInstanceDao.findAll(Example.of(dingTalkProcessInstance),pageRequest);
            if(totalPage==Integer.MAX_VALUE){
                totalPage=pageList.getTotalPages();
            }

            pageList.getContent().forEach(content->{
                if(content.getProcessInstance()==null){
                    log.info("观察异常内容"+content+"");
                }else {
                    if (content.getProcessInstanceId().equals(DingTalkConstant.PROCESS_CODE_MATERIAL_RESEND)){
                        MaterialResendBO materialResendBO = MaterialResendBO.fromDingTak(content.getProcessInstance());
                        try {
                            strings.add(objectMapper.writeValueAsString(materialResendBO));
                        } catch (JsonProcessingException e) {
                            log.error(e.getMessage());
                        }
                    }
                }
            });
            ChuanyunSaveAllDTO chuanyunSaveAllResponse= chuanYunManager.saveAll(MaterialResendBO.SCHEMA_CODE,strings,true);
            log.info("氚云保存物料补发数据结果的错误信息为 "+chuanyunSaveAllResponse.getErrorMessage());
            log.info(chuanyunSaveAllResponse+"");
            pageList.getContent().forEach(content->content.setUpdatedInChuanyun(true));
            log.info("开始更新钉钉审批数据上传状态");
            log.info(dingTalkProcessInstanceDao.saveAll(pageList.getContent())+"");
            log.info("已更新上传状态的钉钉审批数据");
            currentPage++;
        }
    }
}
